@@ -1,10 +1,10 @@
 #
-# Copyright (C) 2015-2018 Joelle Maslak
+# Copyright (C) 2015-2022 Joelle Maslak
 # All Rights Reserved - See License
 #
 
 package Crypt::EAMessage;
-$Crypt::EAMessage::VERSION = '1.006'; # TRIAL
+$Crypt::EAMessage::VERSION = '1.220390'; # TRIAL
 use v5.22;
 
 # ABSTRACT: Simple-to-use Abstraction of Encrypted Authenticated Messages
@@ -51,7 +51,7 @@ around 'BUILDARGS', sub ( $orig, $class, %args ) {
     $class->$orig(%args);
 };
 
-sub _hex_to_raw($hex) {
+sub _hex_to_raw ($hex) {
     $hex =~ s/^0x//;    # Remove 0x leader if it is present
 
     if ( $hex =~ /[^0-9A-Fa-f]/s ) { die("Non-hex characters present in hex_key"); }
@@ -68,7 +68,7 @@ subtype 'Crypt::EAMessage::Key', as 'Str',
   where { _valid_key($_) },
   message { "AES key lengths must be 16, 24, or 32 bytes long" };
 
-sub _valid_key($key) {
+sub _valid_key ($key) {
     my $l = length($_);
 
     if ( ( $l != 16 ) && ( $l != 24 ) && ( $l != 32 ) ) { return; }
@@ -110,19 +110,33 @@ sub encrypt_auth ( $self, $input ) {
 
 
 sub encrypt_auth_ascii ( $self, $input, $eol = undef ) {
-    my $ct = $self->_encrypt_auth_internal($input);
+    my $ct     = $self->_encrypt_auth_internal($input);
     my $base64 = encode_base64( $ct, $eol );
     return "2$base64";    # Type 2 = Base 64
 }
 
-sub _encrypt_auth_internal ( $self, $input ) {
+sub _encrypt_auth_internal ( $self, $input, $opts = {} ) {
     state $random = Bytes::Random::Secure->new( Bits => 1024, NonBlocking => 1 );
+
+    for my $opt ( sort keys %$opts ) {
+        if ( $opt eq 'text' ) { next; }
+
+        die("Unknown option to encrypt: $opt");
+    }
+
     my $nonce = $random->bytes(16);
 
-    my $frozen = nfreeze( \$input );
+    my $data;
+    if ( ( !exists( $opts->{text} ) ) && ( !$opts->{text} ) ) {
+        # Any type of input
+        $data = nfreeze( \$input );
+    } else {
+        # Text only input
+        $data = $input;
+    }
 
     my ( $enc, $tag ) =
-      ccm_encrypt_authenticate( 'AES', $self->raw_key(), $nonce, '', 128, $frozen );
+      ccm_encrypt_authenticate( 'AES', $self->raw_key(), $nonce, '', 128, $data );
 
     my $ct = $nonce . $tag . $enc;
     return $ct;
@@ -139,11 +153,21 @@ sub encrypt_auth_urlsafe ( $self, $input ) {
 }
 
 
+sub encrypt_auth_portable ( $self, $input ) {
+    my $ct = $self->_encrypt_auth_internal( $input, { text => 1 } );
+
+    my $urltext = encode_base64( $ct, "" );
+    $urltext =~ tr|\+/|-_|;
+
+    return "4$urltext";    # Type 3 = Modified Base 64
+}
+
+
 sub decrypt_auth ( $self, $ct ) {
     if ( length($ct) < 34 ) { die("Message too short to be valid") }
 
     my $type = substr( $ct, 0, 1 );
-    my $enc = substr( $ct, 1 );
+    my $enc  = substr( $ct, 1 );
 
     if ( $type eq '1' ) {
         return $self->_decrypt_auth_internal($enc);
@@ -156,13 +180,24 @@ sub decrypt_auth ( $self, $ct ) {
         my $ascii = decode_base64($enc);    # It's okay if this ignores bad base64,
                                             # since we'll fail decryption.
         return $self->_decrypt_auth_internal($ascii);
+    } elsif ( $type eq '4' ) {
+        $enc =~ tr|-_|+/|;
+        my $ascii = decode_base64($enc);    # It's okay if this ignores bad base64,
+                                            # since we'll fail decryption.
+        return $self->_decrypt_auth_internal( $ascii, { text => 1 } );
     } else {
         die("Unsupported encoding type");
     }
 }
 
-sub _decrypt_auth_internal ( $self, $ct ) {
-    if ( length($ct) < 34 ) { die("Message too short to be valid") }
+sub _decrypt_auth_internal ( $self, $ct, $opts = {} ) {
+    if ( length($ct) < 32 ) { die("Message too short to be valid") }
+
+    for my $opt ( sort keys %$opts ) {
+        if ( $opt eq 'text' ) { next; }
+
+        die("Unknown option to decrypt: $opt");
+    }
 
     my $nonce = substr( $ct, 0,  16 );
     my $tag   = substr( $ct, 16, 16 );
@@ -171,8 +206,19 @@ sub _decrypt_auth_internal ( $self, $ct ) {
     my $frozen = ccm_decrypt_verify( 'AES', $self->raw_key(), $nonce, '', $enc, $tag );
     if ( !defined($frozen) ) { die("Could not decrypt message") }
 
-    my $plaintext = thaw($frozen);
-    return $$plaintext;
+    if ( ( !exists( $opts->{text} ) ) && ( !$opts->{text} ) ) {
+        # Perl 5 data structure
+        my $plaintext = thaw($frozen);
+        return $$plaintext;
+    } else {
+        # Plain text
+        return $frozen;
+    }
+}
+
+
+sub generate_key ($self) {
+    return Bytes::Random::Secure::random_bytes_hex(32);
 }
 
 __PACKAGE__->meta->make_immutable;
@@ -191,7 +237,7 @@ Crypt::EAMessage - Simple-to-use Abstraction of Encrypted Authenticated Messages
 
 =head1 VERSION
 
-version 1.006
+version 1.220390
 
 =head1 SYNOPSIS
 
@@ -199,8 +245,10 @@ version 1.006
 
   my $eamsg = Crypt::EAMessage->new( hex_key => $hex );
 
-  $encrypted = $eamsg->encrypt_auth($input);
-  $enc_ascii = $eamsg->encrypt_auth_ascii($input);
+  $encrypted    = $eamsg->encrypt_auth($input);
+  $enc_ascii    = $eamsg->encrypt_auth_ascii($input);
+  $enc_url      = $eamsg->encrypt_auth_urlsafe($input);
+  $enc_portable = $eamsg->encrypt_auth_portable($input); # Input must be text
 
   $decrypted = $eamsg->decrypt_auth($encrypted);
 
@@ -290,8 +338,26 @@ freeze and thaw) passed as a parameter, generating an ASCII (modified
 base64) cipher text output.  This output is safe to pass as part of a
 query string or URL.  Namely, it doesn't use the standard Base 64
 characters C<+> or C</>, replacing them with C<-> and C<_> respectively.
-In addition, rather than the string starting with a "2" (indicating the
-Base 64 variant), it will start with a "3".
+In addition, the cyphertext output will start with a "3" rather than the
+"2" that the base 64 variant starts with.
+
+=head2 encrypt_auth_portable
+
+  my $ciphertext = $ea->encrypt_auth_portable( $plaintext );
+
+Added in version 1.190900
+
+Encrypts the plain text (or byte string) passed as a parameter, generating
+an ASCII (modified base64) cipher text output.  This output is safe to pass
+as part of a query string or URL.  Namely, it doesn't use the standard Base 64
+characters C<+> or C</>, replacing them with C<-> and C<_> respectively.
+In addition, the cyphertext output will start with a "4".
+
+This is intended for cross-language compatibility, so it does not utilize
+store/thaw.
+
+SECURITY NOTE: The contents of a zero length string can be determined from
+the length of the encrypted portable message.
 
 =head2 decrypt_auth
 
@@ -302,21 +368,34 @@ Decrypts the cipher text into the object that was frozen during encryption.
 If the authentication or decryption fails, an exception is thrown. Otherwise
 it returns the plaintext/object.
 
+=head2 generate_key
+
+ say "Hex key: " . Crypt::EAMessage->generate_key()
+
+Added in version 1.220390
+
+This is a class method (I.E. you do not need to instantiate the
+C<Crypt::EAMessage> class to use this).
+
+Returns a randomly generated key suitable to use with AES256 as a hex number.
+
 =head1 GENERATING AES256 KEYS
 
 To generate a key, a simple Perl program can accomplish this - note that you
 should NOT use standard C<rand()> to do this.
 
   use feature 'say';
-  use Bytes::Random::Secure qw(random_bytes_hex);
+  use Crypt::EAMessage;
 
-  my $hexkey = random_bytes_hex(32);
+  my $hexkey = Crypt::EAMessage->generate_key()
   say "Key is: $hexkey";
 
-Alternative, you can do this with a one-liner
-to return a hex key:
+Alternative, you can do this with a one-liner to return a hex key, and the
+L<Crypt::EAMessage::Keygen> module:
 
-  perl -MBytes::Random::Secure=random_bytes_hex -E 'say random_bytes_hex(32)'
+  perl -MCrypt::EAMessage::Keygen -e 1
+
+This will output a random key in hex format suitable for use as an AES256 key.
 
 =head1 SECURITY
 
@@ -356,7 +435,7 @@ Joelle Maslak <jmaslak@antelope.net>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2018 by Joelle Maslak.
+This software is copyright (c) 2019-2022 by Joelle Maslak.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
